Index: src/client.c
===================================================================
--- src/client.c	(.../vendor/4.20)	(revision 12)
+++ src/client.c	(.../branches/4.20-061213)	(revision 12)
@@ -816,7 +816,7 @@
     closesocket(c->fd);
     c->fd=-1; /* avoid double close on cleanup */
     if(strcmp(name, c->opt->username)) {
-        safestring(name);
+        safestring((unsigned char *)name);
         s_log(LOG_WARNING, "Connection from %s REFUSED by IDENT (user %s)",
             c->accepted_address, name);
         longjmp(c->err, 1);
@@ -876,12 +876,12 @@
             if(peer) {
                 safecopy(env[1], "SSL_CLIENT_DN=");
                 X509_NAME_oneline(X509_get_subject_name(peer), name, STRLEN);
-                safestring(name);
+                safestring((unsigned char *)name);
                 safeconcat(env[1], name);
                 putenv(env[1]);
                 safecopy(env[2], "SSL_CLIENT_I_DN=");
                 X509_NAME_oneline(X509_get_issuer_name(peer), name, STRLEN);
-                safestring(name);
+                safestring((unsigned char *)name);
                 safeconcat(env[2], name);
                 putenv(env[2]);
                 X509_free(peer);
@@ -963,20 +963,43 @@
     SOCKADDR_LIST resolved_list, *address_list;
     int error, fd;
     u16 i;
+    char *remote_address = c->opt->remote_address;
     char connecting_address[IPLEN];
 
+#ifdef HAVE_LIBCURL
+    /* If the remote address specifies a URL lookup, then issue it to get the socket specifier. */
+    if (c->opt->remote_address[0] == '@')
+    {
+		X509 *peer = SSL_get_peer_certificate(c->ssl);
+
+        remote_address = url_lookup(peer, &c->opt->remote_address[1], "socket");
+		if (peer)
+			X509_free(peer);
+        c->opt->option.delayed_lookup = 1;
+    }
+#endif
     /* setup address_list */
     if(c->opt->option.delayed_lookup) {
         resolved_list.num=0;
-        if(!name2addrlist(&resolved_list,
-                c->opt->remote_address, DEFAULT_LOOPBACK)){
+        if((remote_address == NULL) || !name2addrlist(&resolved_list, remote_address, DEFAULT_LOOPBACK)){
             s_log(LOG_ERR, "No host resolved");
+#ifdef HAVE_LIBCURL
+            /* Free lookup buffer */
+            if ((c->opt->remote_address[0] == '@') && remote_address)
+                free(remote_address);
+#endif
             longjmp(c->err, 1);
         }
         address_list=&resolved_list;
     } else /* use pre-resolved addresses */
         address_list=&c->opt->remote_addr;
 
+#ifdef HAVE_LIBCURL
+    /* Free lookup buffer */
+    if (c->opt->remote_address[0] == '@')
+        free(remote_address);
+#endif
+
     /* try to connect each host from the list */
     for(i=0; i<address_list->num; i++) {
         memcpy(&addr, address_list->addr + address_list->cur,
Index: src/verify.c
===================================================================
--- src/verify.c	(.../vendor/4.20)	(revision 12)
+++ src/verify.c	(.../branches/4.20-061213)	(revision 12)
@@ -90,6 +90,21 @@
         add_dir_lookup(section->revocation_store, section->ca_dir);
     }
 
+#ifdef HAVE_LIBCURL
+    /* If we have a lookup URL defined, then add a verify handler to process the lookups */
+    if(section->ca_lookup[0])
+    {
+        if(!section->ca_dir)
+        {
+            s_log(LOG_ERR, "CAlookup is configured without a CApath option.");
+            exit(1);
+        }
+        section->ca_lookup[1] = section->ca_dir;
+        SSL_CTX_set_cert_verify_callback(section->ctx, url_ssl_verify, section->ca_lookup);
+        s_log(LOG_DEBUG, "Verify lookup URL set to %s", section->ca_lookup[0]);
+    }
+#endif
+
     if(section->crl_file)
         load_file_lookup(section->revocation_store, section->crl_file);
 
Index: src/stunnel.c
===================================================================
--- src/stunnel.c	(.../vendor/4.20)	(revision 12)
+++ src/stunnel.c	(.../branches/4.20-061213)	(revision 12)
@@ -81,6 +81,7 @@
 void main_initialize(char *arg1, char *arg2) {
     ssl_init(); /* initialize SSL library */
     sthreads_init(); /* initialize critical sections & SSL callbacks */
+    url_init(); /* initialize URL lookup support */
     parse_config(arg1, arg2);
     log_open();
     stunnel_info(0);
Index: src/Makefile.am
===================================================================
--- src/Makefile.am	(.../vendor/4.20)	(revision 12)
+++ src/Makefile.am	(.../branches/4.20-061213)	(revision 12)
@@ -3,7 +3,7 @@
 # File lists
 
 common_headers = common.h prototypes.h
-common_sources = file.c client.c log.c options.c protocol.c network.c resolver.c ssl.c ctx.c verify.c sthreads.c stunnel.c
+common_sources = file.c client.c log.c options.c protocol.c network.c resolver.c ssl.c ctx.c verify.c sthreads.c stunnel.c url_lookup.c
 unix_sources = pty.c
 shared_sources = env.c
 win32_sources = gui.c resources.h resources.rc stunnel.ico
Index: src/url_lookup.c
===================================================================
--- src/url_lookup.c	(.../vendor/4.20)	(revision 0)
+++ src/url_lookup.c	(.../branches/4.20-061213)	(revision 12)
@@ -0,0 +1,379 @@
+/*
+ *   stunnel URL lookup module
+ *
+ *   Copyright (c) 2006 Dan Jakubiec <dan.jakubiec@gmail.com>
+ *                 All Rights Reserved
+ *
+ *   This program is free software; you can redistribute it and/or modify
+ *   it under the terms of the GNU General Public License as published by
+ *   the Free Software Foundation; either version 2 of the License, or
+ *   (at your option) any later version.
+ *
+ *   This program is distributed in the hope that it will be useful,
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *   GNU General Public License for more details.
+ *
+ *   You should have received a copy of the GNU General Public License
+ *   along with this program; if not, write to the Free Software
+ *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *   In addition, as a special exception, Dan Jakubiec gives
+ *   permission to link the code of this program with the OpenSSL
+ *   library (or with modified versions of OpenSSL that use the same
+ *   license as OpenSSL), and distribute linked combinations including
+ *   the two.  You must obey the GNU General Public License in all
+ *   respects for all of the code used other than OpenSSL.  If you modify
+ *   this file, you may extend this exception to your version of the
+ *   file, but you are not obligated to do so.  If you do not wish to
+ *   do so, delete this exception statement from your version.
+ */
+
+#include "common.h"
+#include "prototypes.h"
+
+#ifdef HAVE_LIBCURL
+#include <curl/curl.h>
+
+/* Locals */
+#define URL_LOOKUP_MAXURL               2048
+#define URL_LOOKUP_MAXFIELD             16384
+
+/* Context structure for collecting URL response data */
+typedef struct
+{
+	size_t			max_bytes;
+	char			*buffer;
+} URL_LOOKUP;
+
+/*
+	FUNCTION	url_init
+
+	DESCRIPTION
+		Initializes the URL lookup subsystem.
+
+	PARAMETERS
+		None
+
+	RETURN VALUES
+		None
+*/
+void url_init(void)
+{
+	/* Initialize global libary paramenters */
+	if (curl_global_init(CURL_GLOBAL_NOTHING) != 0)
+	{
+		s_log(LOG_ERR, "Failed to initialize curl URL library");
+		exit(1);
+	}
+}
+
+/*
+	FUNCTION	url_getdata
+
+	DESCRIPTION
+		libcurl callback function for aggregating data received via the URL.
+		This function receives only the body content, not protocol headers.
+
+	PARAMETERS
+		buffer - Buffer containing data obtained via the URL
+		size - Size of each member
+		nmemb - Number of members in buffer
+		userp - Pointer to a URL_LOOKUP structure
+
+	RETURN VALUES
+		Number of bytes accepted
+*/
+static size_t url_getdata(void *buffer, size_t size, size_t nmemb, void *userp)
+{
+	URL_LOOKUP *lookup = (URL_LOOKUP *)userp;
+	size_t rc = size * nmemb;
+	size_t num_bytes = (rc < lookup->max_bytes)?rc:lookup->max_bytes;
+	char *url_body = (char *)buffer;
+
+	/* Copy payload bytes received from the URL server until we fill up our buffer */
+	while (num_bytes-- > 0)
+	{
+		*lookup->buffer++ = *url_body++;
+		lookup->max_bytes--;
+	}
+	return rc;
+}
+
+/*
+	FUNCTION	url_sprintf
+
+	DESCRIPTION
+		Does sprintf-style parameter replacement for a URL lookup template.  The
+		following parameters are supported:
+
+			%h - Replaced with the OpenSSL hash of the peer cert's subject line
+			%n - Replaced with the common name found in the peer cert's subject line
+
+	PARAMETERS
+		peer - Peer certificate
+		url - Buffer where resulting URL string will be written (must be URL_LOOKUP_MAXURL bytes large)
+		url_format - The URL template string
+
+	RETURN VALUES
+		Return url on success, or NULL if the resulting URL is too long
+*/
+static char *url_sprintf(X509 *peer, char *url, const char *url_format)
+{
+	unsigned int i;
+	int num_bytes;
+	char *buffer, *s;
+
+	/* Replace any variables present in the url_format string.  Variables are embedded
+	 * in the URL format string using %<letter> specifiers.  To prevent conflicts with
+	 * URL-encoded %HH sequences, <letter> must be something other than a hexadecimal
+	 * digit.
+	 */
+	for (i = 0; *url_format && (i < URL_LOOKUP_MAXURL); url_format++)
+	{
+		if (*url_format == '%')
+		{
+			/* Variable specifier.  Determine which URL variable to insert.
+			 * (WARNING: hexadecimal digits are reserved for URL encoding!)
+			 */
+			switch (url_format[1])
+			{
+				case 'h':   /* Client certificate hash */
+					url_format++;
+					num_bytes = 0;
+					if (peer && ((URL_LOOKUP_MAXURL - i) > 8))
+					{
+						sprintf(&url[i], "%08lx", X509_NAME_hash(X509_get_subject_name(peer)));
+						num_bytes = 8;
+						i += 8;
+					}
+					if (num_bytes <= 0)
+						s_log(LOG_WARNING, "URL lookup was unable to extract the client certificate's hash.");
+					continue;
+
+				case 'n':   /* Common name from client certificate subject */
+					url_format++;
+					if (peer)
+					{
+						/* Get the subject name from the certificate */
+						num_bytes = 0;
+						buffer = X509_NAME_oneline(X509_get_subject_name(peer), NULL, 0);
+						if (buffer)
+						{
+							char *common_name = strstr(buffer, "/CN=");
+
+							/* Find the common name in the subject.  It should be formatted as
+							 *      [...]/CN=<name>[/...]
+							 */
+							if (common_name)
+							{
+								common_name += 4;
+
+								/* Find the end of the common name: either the next slash or the
+								 * end of the string.
+								 */
+								s = strchr(common_name, '/');
+								if (s)
+									num_bytes = s - common_name;
+								else
+									num_bytes = strlen(common_name);
+
+								/* URL-encode the name */
+								s = curl_escape(common_name, num_bytes);
+								if (s)
+								{
+									num_bytes = strlen(s);
+									if (num_bytes < (URL_LOOKUP_MAXURL - i))
+									{
+										memcpy(&url[i], s, num_bytes);
+										i += num_bytes;
+									} else
+										num_bytes = 0;
+									curl_free(s);
+								} else
+									num_bytes = 0;
+							}
+							OPENSSL_free(buffer);
+						}
+						if (num_bytes == 0)
+							s_log(LOG_WARNING, "URL lookup was unable to extract the client certificate's common name.");
+					} else
+						s_log(LOG_WARNING, "URL lookup wanted a client certificate, but none was provided.");
+					continue;
+
+				default:
+					/* Unrecognized escape sequence.  Hopefully this is a URL encoding 
+					 * sequence.  Just copy the bytes over.
+					 */
+					break;
+			}
+		}
+		url[i++] = *url_format;
+	}
+	if (i >= URL_LOOKUP_MAXURL)
+	{
+		s_log(LOG_ERR, "URL was too long (exceeded %u bytes).", URL_LOOKUP_MAXURL);
+		return NULL;
+	}
+	url[i] = '\0';
+	return url;
+}
+
+/*
+	FUNCTION	url_lookup
+
+	DESCRIPTION
+		Issue a query to provided URL to retrieve information for the specified
+		field.
+
+	PARAMETERS
+		peer - Peer certificate
+		url_format - URL to query.  May contain certain % parameters for replacement.
+		field_name - The URL response is expected to begin with "<field_name>="
+
+	RETURN VALUES
+		malloc'd buffer containing URL response, or NULL on error.  Caller is responsible
+		for freeing the returned buffer.
+*/
+char *url_lookup(X509 *peer, const char *url_format, const char *field_name)
+{
+	CURL *hCurl;
+	size_t field_len = strlen(field_name);
+	char *buffer;
+	char *url;
+	URL_LOOKUP lookup;
+
+	/* Allocate a buffer to store URL and field information */
+	buffer = malloc(URL_LOOKUP_MAXURL + URL_LOOKUP_MAXFIELD);
+	if (buffer == NULL)
+	{
+		s_log(LOG_ERR, "Out of memory");
+		return NULL;
+	}
+
+	/* Generate the URL string, replacing any variables with appropriate values. */
+	url = url_sprintf(peer, &buffer[URL_LOOKUP_MAXFIELD], url_format);
+	if (url == NULL)
+	{
+		free(buffer);
+		return NULL;
+	}
+	s_log(LOG_DEBUG, "Doing URL lookup to '%s' for '%s'.", url, field_name);
+
+	/* Allocate CURL library resources */
+	hCurl = curl_easy_init();
+	if (!hCurl)
+	{
+		s_log(LOG_ERR, "curl_easy_init() failed");
+		free(buffer);
+		return NULL;
+	}
+
+	/* Setup a control buffer to collect the server response */
+	lookup.max_bytes = URL_LOOKUP_MAXFIELD - 1;
+	lookup.buffer = buffer;
+	if ((curl_easy_setopt(hCurl, CURLOPT_URL, url) == CURLE_OK) &&
+		(curl_easy_setopt(hCurl, CURLOPT_WRITEFUNCTION, url_getdata) == CURLE_OK) &&
+		(curl_easy_setopt(hCurl, CURLOPT_WRITEDATA, &lookup) == CURLE_OK))
+	{
+		/* Process the URL request */
+		curl_easy_perform(hCurl);
+		*lookup.buffer = '\0';
+
+		/* A valid response must begin with '<field_name>='.  If we got this, then remove the
+		 * field prefix and return the socket specifier.
+		 */
+		if ((strncmp(buffer, field_name, field_len) == 0) && (buffer[field_len] == '='))
+		{
+			strcpy(&buffer[0], &buffer[field_len + 1]);
+			s_log(LOG_INFO, "URL lookup for '%s' returned '%s'.", field_name, buffer);
+		} else
+			*buffer = '\0';
+	} else
+		*buffer = '\0';
+
+	/* Cleanup */
+	curl_easy_cleanup(hCurl);
+	if (buffer[0] == '\0')
+	{
+		s_log(LOG_ERR, "URL lookup to '%s' for '%s' failed.", url, field_name);
+		free(buffer);
+		return NULL;
+	}
+	return buffer;
+}
+
+/*
+	FUNCTION	url_ssl_verify
+
+	DESCRIPTION
+		OpenSSL verify callback.  This function is used to dynamically lookup and load
+		new CA certificates based on the certificate information provided by the peer.
+
+	PARAMETERS
+		store_ctx - Certificate store context
+		context - Pointer to char *[2].  First string is the lookup URL, the second is the CA directory.
+
+	RETURN VALUES
+		1 on verify OK, 0 on failure
+*/
+int url_ssl_verify(X509_STORE_CTX *store_ctx, void *context)
+{
+	const char *ca_lookup = ((char **)context)[0];
+	const char *ca_dir = ((char **)context)[1];
+	FILE *fp;
+	char *pem;
+	int rc;
+	char cert_fn[256];
+
+	/* Issue a URL lookup to get the client certificate */
+	cert_fn[0] = '\0';
+	pem = url_lookup(store_ctx->cert, ca_lookup, "certificate");
+	if (pem)
+	{
+		/* Write the certificate to the appropriate file in the CA directory.  Not that the server may have
+		 * multiple certificates for the requested hash value.  That should be okay as long as all the
+		 * certs are returned.  The resulting PEM file should be read by the SSL library and all the
+		 * cert should be loaded.
+		 *
+		 * Any existing certificate file with the same hash is overwritten.  OpenSSL should re-read this
+		 * file on every handshake.
+		 */
+		snprintf(cert_fn, sizeof(cert_fn), "%s/%08lx.0", ca_dir, X509_NAME_hash(X509_get_subject_name(store_ctx->cert)));
+		fp = fopen(cert_fn, "w");
+		if (fp)
+		{
+			if (fwrite(pem, strlen(pem), 1, fp) != 1)
+			{
+				fclose(fp);
+				unlink(cert_fn);
+				cert_fn[0] = '\0';
+			} else
+				fclose(fp);
+		} else
+			cert_fn[0] = '\0';
+		free(pem);
+	}
+
+	/* Issue a warning if the certificate lookup failed, but go ahead and run the certificate verify
+	 * anyway.  There may be other certificate sources which match the provided cert.
+	 */
+	if (cert_fn[0] == '\0')
+		s_log(LOG_WARNING, "URL certificate lookup failed.");
+
+	/* Call the normal OpenSSL certificate verification function */
+	rc = X509_verify_cert(store_ctx);
+
+	/* Remove the certificate file.  We only want to use it once. */
+	if (cert_fn[0] != '\0')
+		unlink(cert_fn);
+
+	/* Return the verification result */
+	return rc;
+}
+
+#else /* HAVE_LIBCURL */
+void url_init(void)
+{
+}
+#endif /* HAVE_LIBCURL */
Index: src/options.c
===================================================================
--- src/options.c	(.../vendor/4.20)	(revision 12)
+++ src/options.c	(.../branches/4.20-061213)	(revision 12)
@@ -502,6 +502,29 @@
         break;
     }
 
+#ifdef HAVE_LIBCURL
+    /* CAlookup */
+    switch(cmd) {
+    case CMD_INIT:
+        section->ca_lookup[0]=section->ca_lookup[1]=NULL;
+        break;
+    case CMD_EXEC:
+        if(strcasecmp(opt, "CAlookup"))
+            break;
+        if(arg[0]) /* not empty */
+            section->ca_lookup[0]=stralloc(arg);
+        else
+            section->ca_lookup[0]=NULL;
+        return NULL; /* OK */
+    case CMD_DEFAULT:
+        break;
+    case CMD_HELP:
+        log_raw("%-15s = CA certificate lookup URL for 'verify' option",
+            "CAlookup");
+        break;
+    }
+#endif /* HAVE_LIBCURL */
+
     /* CAfile */
     switch(cmd) {
     case CMD_INIT:
@@ -606,6 +629,14 @@
             break;
         section->option.remote=1;
         section->remote_address=stralloc(arg);
+#ifdef HAVE_LIBCURL
+        /* If the socket information is to be obtained via URL lookup, then force
+         * delayed lookup.  URL lookups generally require information from
+         * the connecting client.
+         */
+        if(arg[0] == '@')
+            section->option.delayed_lookup=1;
+#endif
         if(!section->option.delayed_lookup &&
                 !name2addrlist(&section->remote_addr, arg, DEFAULT_LOOPBACK)) {
             log_raw("Cannot resolve '%s' - delaying DNS lookup", arg);
Index: src/prototypes.h
===================================================================
--- src/prototypes.h	(.../vendor/4.20)	(revision 12)
+++ src/prototypes.h	(.../branches/4.20-061213)	(revision 12)
@@ -163,6 +163,9 @@
 
         /* service-specific data for ctx.c */
     char *ca_dir;                              /* directory for hashed certs */
+#ifdef HAVE_LIBCURL
+    char *ca_lookup[2];        /* lookup URL for dynamic certificate loading */
+#endif
     char *ca_file;                       /* file containing bunches of certs */
     char *crl_dir;                              /* directory for hashed CRLs */
     char *crl_file;                       /* file containing bunches of CRLs */
@@ -436,6 +439,12 @@
 LPSTR tstr2str(const LPTSTR);
 #endif
 
+extern void url_init(void);
+#ifdef HAVE_LIBCURL
+extern char *url_lookup(X509 *peer, const char *uri, const char *field_name);
+extern int url_ssl_verify(X509_STORE_CTX *store, void *context);
+#endif
+
 #endif /* defined PROTOTYPES_H */
 
 /* End of prototypes.h */
Index: configure.ac
===================================================================
--- configure.ac	(.../vendor/4.20)	(revision 12)
+++ configure.ac	(.../branches/4.20-061213)	(revision 12)
@@ -103,6 +103,15 @@
 then LIBS="$LIBS -lbe -lroot -lbind"
 fi
 
+# Add libraries needed for URL lookup
+AC_CHECK_LIB(curl, curl_global_init,
+        [
+            LIBS="$LIBS `curl-config --libs`"
+            HAVE_LIBCURL="yes"
+            AC_DEFINE(HAVE_LIBCURL)
+        ]
+    )
+
 AC_MSG_NOTICE([**************************************** thread model])
 
 checkpthreadlib() { :
Index: doc/stunnel.pod
===================================================================
--- doc/stunnel.pod	(.../vendor/4.20)	(revision 12)
+++ doc/stunnel.pod	(.../branches/4.20-061213)	(revision 12)
@@ -291,6 +291,13 @@
 
 I<CApath> path is relative to I<chroot> directory if specified.
 
+=item B<CAlookup> = lookup_url
+
+If CApath is specified, you may use this option to specify a lookup URL for obtaining
+additional CA certificates.
+
+See the I<DYNAMIC URL LOOKUP> section for more information on specifying URL lookups.
+
 =item B<CAfile> = certfile
 
 Certificate Authority file
@@ -320,12 +327,14 @@
 
 default: no (server mode)
 
-=item B<connect> = [host:]port
+=item B<connect> = [host:]port | @url_lookup
 
-connect to remote host:port
+connect to remote host:port, or get the host:port information from the specified url
 
 If no host specified, defaults to localhost.
 
+See the I<DYNAMIC URL LOOKUP> section for more information on specifying URL lookups.
+
 =item B<CRLpath> = directory
 
 Certificate Revocation Lists directory
@@ -506,6 +515,25 @@
 B<stunnel> returns zero on success, non-zero on error.
 
 
+=head1 DYNAMIC URL LOOKUPS
+
+Certain configuration parameters may be obtained dynamically via URL lookups.
+When specifying a lookup URL, any URL type support by the libcurl URL library
+may be used (e.g. "http://...", "ftp://...", etc.).
+
+The URL may contain the format specifiers described below:
+
+	%n - replaced with the "common name" found in the peer certificate
+
+	%h - replaced with the hash ID of the peer certificate
+
+The URL response must return one of the following items, formatted as follows:
+
+	socket=<host>:<port>
+
+	certificate=<PEM formatted file>
+
+
 =head1 EXAMPLES
 
 In order to provide SSL encapsulation to your local I<imapd> service, use
@@ -546,7 +574,15 @@
     engineNum=1
     key=id_45
 
+To enable dynamic URL lookups, use something like
 
+    CApath = /var/ca_certs
+    CAlookup = http://server/lookup.cgi?hash=%h
+
+    [service]
+    connect = @http://server/lookup.cgi?name=%n
+
+
 =head1 FILES
 
 =over 4
